// Copyright 2009, 2010 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.tapestry5.internal.transform;

import org.apache.tapestry5.annotations.SessionAttribute;
import org.apache.tapestry5.ioc.services.FieldValueConduit;
import org.apache.tapestry5.model.MutableComponentModel;
import org.apache.tapestry5.services.ClassTransformation;
import org.apache.tapestry5.services.ComponentClassTransformWorker;
import org.apache.tapestry5.services.Request;
import org.apache.tapestry5.services.Session;
import org.apache.tapestry5.services.TransformField;

/**
 * Looks for the {@link SessionAttribute} annotation and converts read and write access on such
 * fields into calls to the {@link Session#getAttribute(String)} and {@link Session#setAttribute(String, Object)}.
 */
public class SessionAttributeWorker implements ComponentClassTransformWorker
{
    private final Request request;

    private class SessionKeyConduit implements FieldValueConduit
    {
        private final String key;

        public SessionKeyConduit(String key)
        {
            this.key = key;
        }

        private Session getSession()
        {
            return request.getSession(true);
        }

        public Object get()
        {
            // TODO: caching, and not creating the session unnecessarily
            return getSession().getAttribute(key);
        }

        public void set(Object newValue)
        {
            getSession().setAttribute(key, newValue);
        }
    }

    public SessionAttributeWorker(Request request)
    {
        this.request = request;
    }

    public void transform(ClassTransformation transformation, MutableComponentModel model)
    {
        for (TransformField field : transformation.matchFieldsWithAnnotation(SessionAttribute.class))
        {
            convertFieldToSessionAccess(field);
        }
    }

    private void convertFieldToSessionAccess(TransformField field)
    {
        SessionAttribute annotation = field.getAnnotation(SessionAttribute.class);

        field.claim(annotation);

        String key = determineSessionKey(field, annotation.value());

        field.replaceAccess(new SessionKeyConduit(key));
    }

    private String determineSessionKey(TransformField field, String value)
    {
        return value.equals("") ? field.getName() : value;
    }
}
